import sys, os, re
import pprint

sys.path.append("../src")
import pronlex
import phonecodes
from sentences import sents

"""This script should be run from the command line, in the test directory.
   It will load some pronlexes from the 'fixtures' subdirectory,
   test phone code conversion, and test both word and phone searches."""

os.makedirs("outputs", exist_ok=True)
# First, test the phonecode conversions
with open("outputs/test_phonecodes.txt", "w") as f:
    original = sys.stdout
    sys.stdout = f
    phonecode_cases = [
        ("arpabet", "ipa", phonecodes.arpabet2ipa, "eng"),
        ("ipa", "arpabet", phonecodes.ipa2arpabet, "eng"),
        ("ipa", "callhome", phonecodes.ipa2callhome, "arz"),
        ("ipa", "callhome", phonecodes.ipa2callhome, "cmn"),
        ("ipa", "callhome", phonecodes.ipa2callhome, "spa"),
        ("callhome", "ipa", phonecodes.callhome2ipa, "arz"),
        ("callhome", "ipa", phonecodes.callhome2ipa, "cmn"),
        ("callhome", "ipa", phonecodes.callhome2ipa, "spa"),
        ("ipa", "disc", phonecodes.ipa2disc, "deu"),
        ("ipa", "disc", phonecodes.ipa2disc, "eng"),
        ("ipa", "disc", phonecodes.ipa2disc, "nld"),
        ("disc", "ipa", phonecodes.disc2ipa, "deu"),
        ("disc", "ipa", phonecodes.disc2ipa, "eng"),
        ("disc", "ipa", phonecodes.disc2ipa, "nld"),
        ("ipa", "xsampa", phonecodes.ipa2xsampa, "amh"),
        ("ipa", "xsampa", phonecodes.ipa2xsampa, "ben"),
        ("xsampa", "ipa", phonecodes.xsampa2ipa, "amh"),
        ("xsampa", "ipa", phonecodes.xsampa2ipa, "ben"),
    ]
    for case in phonecode_cases:
        print("##### %s2%s ############################" % (case[0], case[1]))
        print("%s %s" % (case[0], sents[case[3]][case[0]]))
        res = case[2](sents[case[3]][case[0]], case[3])
        suc = "Pass" if res == sents[case[3]][case[1]] else "FAIL"
        print("%s: %s *** %s" % (case[1], res, suc))
        print("")
    sys.stdout = original



from typing import Dict, List, Optional, Tuple, Union
from unicodedata import normalize
import numpy as np


class PhoneFeature:
    def __init__(self, feature: str, value: float) -> None:
        """
        Create a new instance of a ``PhoneFeature`` with the feature and value provided.

        Args:
            feature: The feature to be evaluated.
            value: The value of the feature.
        """
        self.feature = feature.lower()
        self.value = value


class Phone:
    def __init__(
        self,
        index: str,
        features: Dict[str, Union[int, str]],
        language_code: Optional[str] = None,
        allophones: Optional[List[str]] = None,
        collection: Optional[object] = None,
    ) -> None:
        """Create a new `Phone` object.

        Args:
            index: The index of the phone in the phone set.
            features: A dictionary of features. When not provided with numerical values, `-` will be replaced with `-1` and `+` with `1`.
            language_code: The language code of the language that the phoneme belongs to.
            allophones: A list of allophones for the phoneme.
            collection: The ``PhoneCollection`` the phone is contained in.

        Example:
            The ``Phone`` class supports arithmetic operations.
            ```py
            pc = PhoneCollection()
            pc.phones("i").val + pc.phones("u").val
            ```
            > ``[(0.7071067811865476, iu (adn)),(0.7071067811865476, iu (bhg)),...]``

            And data augmentation.
            ```py
            pc = PhoneCollection()
            z = pc.phones("z").val
            z_noise = pc.phones("z").val.noise(.05, random_state=42)
            z.vector - z_noise.vector
            ```
            > ``array([ 0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,``
            > ``        0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,  0.,``
            > ``        0.,  0.,  0., -2.,  0.,  0.,  0.,  0.,  0.,  0.,  0.])``

            If the phones vector has been altered, we can also find its closest existing phone(s).
            ```py
            pc = PhoneCollection()
            z_noise = pc.phones("z").val.noise(.05, random_state=42).closest()
            ```
            > ``[(0.0, z̤ (xho))]``

            \... filtered by language(s)
            ```py
            pc = PhoneCollection()
            z_noise = pc.phones("z").val.noise(.05, random_state=42).langs("eng").closest()
            ```
            > ``[(2.0, z (eng))]``
        """
        self.index = normalize("NFC", index)
        self.feature_names = [f.lower() for f in sorted(features.keys())]
        features = {k.lower(): v for k, v in features.items()}
        self.features = features
        self.vector = [features[k] for k in self.feature_names]
        self.vector = np.array(self.vector)
        self.language_code = language_code
        if not isinstance(allophones, list):
            try:
                self.allophones = [a for a in allophones.split() if a != index]
            except AttributeError:
                # NaN
                self.allophones = []
        else:
            self.allophones = allophones
        self.collection = collection

    def get_feature_vector(self, features: List[str]) -> np.ndarray:
        """
        Get the feature vector of the phone for the features provided. 

        Returns:
            A numpy array of the feature vector.
        """
        return np.array([self.features[k] for k in self.feature_names if k in features])

    @staticmethod
    def _normalize(v) -> List[float]:
        v[v > 1] = -1
        v[v < -1] = 1
        return v.round()

    def __repr__(self) -> str:
        result = self.index
        if self.language_code is not None:
            result += f" ({self.language_code})"
        return result

    def __eq__(self, __o: object) -> bool:
        return np.allclose(self.vector, __o.vector)

    def __add__(self, __o: object) -> object:
        assert self.collection is not None
        vector = self.vector
        if isinstance(__o, PhoneFeature):
            vector[self.feature_names.index(__o.feature)] += __o.value
        else:
            assert self.collection.lang_filter == __o.collection.lang_filter
            vector += __o.vector
        return self.collection.get_closest_by_vector(np.clip(vector, -1, 1))

    def __sub__(self, __o: object) -> object:
        assert self.collection is not None
        vector = self.vector
        if isinstance(__o, PhoneFeature):
            vector[self.feature_names.index(__o.feature)] -= __o.value
        else:
            assert self.collection.lang_filter == __o.collection.lang_filter
            vector -= __o.vector
        return self.collection.get_closest_by_vector(np.clip(vector, -1, 1))

    def __lt__(self, __o: object) -> bool:
        return self.index < __o.index

    def langs(self, langs: str) -> object:
        """
        The langs function takes a string or list of languages and returns the phone with the languages
        filter applied.

        Args:
            langs: A list of language codes or single language code.

        Returns:
            A new instance of the class.
        """
        self.collection = self.collection.langs(langs)
        return self

    def noise(
        self,
        p: float = 0.005,
        abs_max_change: float = 2,
        return_close=False,
        random_state: int = None,
    ) -> Union[List[Tuple[float, object]], object]:
        """
        Given a phone, it will return a new phone with a random vector that is close to the original phone.

        Args:
            p: The element-wise probability of a change in a phone vector.
            abs_max_change: The maximum absolute value an element of the phone vector can change.
            random_state: Seed used for the random numbers used.

        Returns:
            The phone object with a noised feature vector.
        """
        if random_state is not None:
            np.random.seed(random_state)
        probs = np.random.uniform(size=len(self.vector))
        self.vector[probs <= p] += np.random.uniform(
            -abs_max_change, abs_max_change, size=len(self.vector)
        )[probs <= p]
        self.vector = Phone._normalize(self.vector)
        return self

    def closest(self) -> List[Tuple[float, object]]:
        """
        Given the current phone's vector, return the closest phone(s) in the collection and their distances.

        Returns:
            A list of distance,phone tuples.
        """
        return self.collection.get_closest_by_vector(self.vector)



import sys, os, re
import phonecodes
import phonecode_tables

LEAF = "LEAF"
MAXSET = 8000


def dict_enter_word(d, w, p):
    if len(p) == 0:
        if LEAF not in d:
            d[LEAF] = set()
        d[LEAF].add(w)
    else:
        if p[0] not in d:
            d[p[0]] = {}
        dict_enter_word(d[p[0]], w, p[1:])


def dict_convert_phonecode(d0, code0, code1, language):
    d1 = dict()
    for p0 in d0:
        if p0 == LEAF:
            d1[p0] = d0[p0]
        else:
            p1 = phonecodes.convert(p0, code0, code1, language)
            d1[p1] = dict_convert_phonecode(d0[p0], code0, code1, language)
    return d1


def dict_phones2words(d, rtd, ph, dist=0, paths=[[]], nmax=MAXSET):
    """Return up to nmax wordstrings, in a list, that match ph
    with up to `dist' distance."""
    res = []
    # If we're at the root node, with no ph, then current paths is a result
    # Prepend each path with the remaining dist
    if d == rtd and len(ph) == 0:
        res.extend([[-dist] + x for x in paths])
    # If the next phone can extend current word, then do it
    if len(ph) > 0 and ph[0] in d:
        res.extend(dict_phones2words(d[ph[0]], rtd, ph[1:], dist, paths, nmax))
    # If we've finished any word, then try to start a new one
    if d != rtd and LEAF in d:
        for w in d[LEAF]:
            ws = [x + [w] for x in paths]
            r = dict_phones2words(rtd, rtd, ph, dist, ws, nmax)
            while len(res) < nmax and len(r) > 0:
                res.append(r.pop())
    # If we can sub the next phone to extend current word, then do it
    if len(ph) > 0 and ph[0] not in d and dist > 0:
        for p in d:
            if p != LEAF:
                res.extend(dict_phones2words(d[p], rtd, ph[1:], dist - 1, paths, nmax))
    # If we can extend current word by skipping a dict phone, then do it
    if dist > 0:
        for p in d:
            if p != LEAF:
                res.extend(dict_phones2words(d[p], rtd, ph, dist - 1, paths, nmax))
    # If we can extend current word by skipping a listed phone, then do it
    if len(ph) > 1 and ph[1] in d and dist > 0:
        res.extend(dict_phones2words(d[ph[1]], rtd, ph[2:], dist - 1, paths, nmax))
    # return up to nmax results
    res = sorted(res)
    return res[:nmax]


###########################################################
def read_raw_dictfile(filename, language):
    """Read from unformatted dictfile: eliminate extra spaces and parens,
    insert a tab.  Return a list of word,pron pairs."""
    S = []
    with open(filename) as f:
        for line in f:
            words = re.split(r"\s+", line.rstrip())
            S.append((words[0], words[1:]))
    return S


def read_isle_dictfile(filename, language, params):
    """Read from ISLE-formatted dictfile, which is just IPA pronunciations,
    but word has parentheses and part of speech to be removed.
    If params['discard_phones'] is a set of chars, then any phones in the set
    will be discarded.  For example, set('#.') discards morph, syl boundaries.
    If params['discard_diacritics'] is a strings, then chars in the string
    will be discarded.  For example, 'ˈˌ' discards stress markers.
    """
    S = []
    with open(filename) as f:
        for line in f:
            txt = re.sub(r"\([^\)]*\)", "", line.rstrip())
            if "discard_diacritics" in params:
                patstr = r"[%s]+" % (params["discard_diacritics"])
                txt = re.sub(patstr, "", txt)
            words = re.split(r"\s+", txt)
            if "discard_phones" in params:
                pr = [p for p in words[1:] if p not in params["discard_phones"]]
            else:
                pr = words[1:]
            S.append((words[0], pr))
    return S


_babel_pcols = {
    "amh": 2,
    "asm": 2,
    "ben": 2,
    "yue": 2,
    "ceb": 1,
    "luo": 1,
    "kat": 2,
    "gug": 1,
    "hat": 1,
    "ibo": 1,
    "jav": 1,
    "kur": 1,
    "lao": 2,
    "lit": 1,
    "mon": 2,
    "pus": 2,
    "swa": 1,
    "tgl": 1,
    "tam": 2,
    "tpi": 1,
    "tur": 1,
    "vie": 1,
    "zul": 1,
}


def read_babel_dictfile(filename, lang, params):
    """Read from a Babel dictfile in language lang.
    If params['pcol'], then it specifies which column contains phones.
    Otherwise, try to guess based on language.
    """
    S = []
    if "pcol" in params:
        pcol = params["pcol"]
    else:
        pcol = _babel_pcols[lang]
    with open(filename) as f:
        for line in f:
            words = re.split(r"\t", line.rstrip(), pcol)
            options = words[pcol].split("\s+")
            for option in options:
                S.append((words[0], option.split()))
    return S


###########################################################
_celex_pcols = {"eng": 6, "deu": 4, "nld": 4}


def read_celex_dictfile(filename, lang, params):
    """Read from a CELEX dictfile in language lang.'
    If params['pcol'] exists, it should be an integer,
    specifying which column contains the phonemes.
    """
    S = []
    if "pcol" in params:
        pcol = params["pcol"]
    else:
        pcol = _celex_pcols[lang]
    with open(filename) as f:
        for line in f:
            words = re.sub(r"-", "", line).rstrip().split("\\")
            word = re.sub(r"\s+", "_", words[1])
            options = words[pcol].split("\\")
            for option in options:
                ol = phonecodes.attach_tones_to_vowels(
                    list(option), set("'"), phonecode_tables._disc_vowels, 1, -1
                )
                S.append((word, ol))
    return S


###################################################################
# _callhome_columns={'arz':(1,2,3,r'([@aiu%AIOUE])'),'cmn':(0,3,2,r'([iI%eEU&a@o>uR])'),'spa':(0,2,3,r'([aeiou]+)')}
_callhome_columns = {"arz": (1, 2, 3), "cmn": (0, 3, 2), "spa": (0, 2, 3)}


def read_callhome_dictfile(filename, L, params):
    """Read a callhome dictionary in language L.
    If params['callhome_columns'] exists, then it should be a tuple
    (grapheme column, phone column, tone column).
    """
    S = []
    if "callhome_columns" in params:
        (gcol, pcol, tcol) = params["callhome_columns"]
    else:
        (gcol, pcol, tcol) = _callhome_columns[L]
    with open(filename) as f:
        for line in f:
            fields = line.rstrip().split("\t")  # Split fields by tab
            tone_options = fields[tcol].split("//")  # split options by slash
            syl_options = fields[pcol].split("//")
            if len(tone_options) == 1:
                options = [(tone_options[0], s) for s in syl_options]
            elif len(tone_options) == len(syl_options):
                options = [
                    (tone_options[n], syl_options[n])
                    for n in range(0, len(syl_options))
                ]
            else:
                raise KeyError(
                    "tone_options and syl_options mismatched: {}".format(line.rstrip())
                )
            for option in options:
                syls = option[1].split(" ")
                tones = list(option[0])
                phones = [syls[n] + tones[n] for n in range(0, len(tones))]
                phones.append("".join(syls[len(tones) :]))
                il = list("".join(phones))
                ol = phonecodes.attach_tones_to_vowels(
                    il,
                    set(phonecode_tables._tone2ipa[L].keys()),
                    phonecode_tables._callhome_vowels[L],
                    -1,
                    1,
                )
                S.append((fields[gcol], ol))
    return S


###########################################################################
def write_dictfile(w2p, filename):
    """Write a w2p dictionary to filename"""
    lines2write = set()  # make sure each written line is unique
    for (w, pron) in w2p.items():
        if len(w) > 0 and len(pron) > 0:
            pl = [
                re.sub(r"\s*Í¡\s*", "", p)
                for p in pron
                if not p.isspace() and len(p) > 0
            ]
            pstr = " ".join(pl)
            if len(pstr) > 0:
                lines2write.add("%s\t%s\n" % (w, pstr))
    with open(filename, "w") as f:
        for line in sorted(lines2write):
            f.write(line)


###################################################################
_dictread = {
    "babel": read_babel_dictfile,
    "callhome": read_callhome_dictfile,
    "celex": read_celex_dictfile,
    "isle": read_isle_dictfile,
}


class lex:
    def __init__(self, language, phonecode="ipa"):
        self.p2w = {}
        self.w2p = {}
        self.language = language
        self.phonecode = phonecode

    def copy(self):
        other = lex(self.language, self.phonecode)
        other.p2w.update(self.p2w)
        other.w2p.update(self.w2p)
        return other

    def add(self, word, pron):
        dict_enter_word(self.p2w, word, pron)
        self.w2p[word] = pron
        return self

    def recode(self, code1):
        if self.phonecode == code1:
            return self.copy()
        else:
            other = lex(self.language, code1)
            other.p2w = dict_convert_phonecode(
                self.p2w, self.phonecode, code1, self.language
            )
            for (w, p) in self.w2p.items():
                q = phonecodes.convertlist(p, self.phonecode, code1, self.language)
                other.w2p[w] = q
            return other

    def phones2words(self, ph, D=0, paths=[[]], nmax=MAXSET):
        """Find all word sequences matching ph with up to D string edit dist.
        Return: a dict mapping distances to lists of lists:
        L=lex()
        res=L.phones2words(ph, 2)
        sntce in res[0] is a list of words matching ph with 0 edit distance
        sntce in res[1] is a list of words matching ph with 1 edit distance
        sntce in res[2] is a list of words matching ph with 2 edit distance
        If the total number of candidates is greater than nmax,
        then the set is truncated at nmax candidates, with a priority
        queue behavior, i.e., the highest-edit-distance candidates
        are discarded first."""
        retval = {}
        listlist = dict_phones2words(self.p2w, self.p2w, ph, D, paths, nmax)
        for res in listlist:
            if len(res) > 1:
                if res[0] + D not in retval:
                    retval[res[0] + D] = []
                retval[res[0] + D].append(res[1:])
        return retval

    def words2phones(self, words):
        pron = []
        for w in words:
            if w in self.w2p:
                pron.extend(self.w2p[w])
            elif w.upper() in self.w2p:
                pron.extend(self.w2p[w.upper()])
            elif w.lower() in self.w2p:
                pron.extend(self.w2p[w.lower()])
            elif w.capitalize() in self.w2p:
                pron.extend(self.w2p[w.capitalize()])
            else:
                pron.append(w)
        return pron

    def read(self, filename, lang, dicttype, params={}):
        if dicttype in _dictread:
            plist = _dictread[dicttype](filename, lang, params)
            for p in plist:
                self.add(p[0], p[1])
        else:
            plist = read_raw_dictfile(filename, lang)
            for p in plist:
                self.add(p[0], p[1])
        return self

    def save(self, filename):
        write_dictfile(self.w2p, filename)


_dict2phonecode = {"babel": "xsampa", "celex": "disc", "isle": "ipa"}


def read(filename, lang, dicttype, params={}):
    if dicttype in _dict2phonecode:
        newlex = lex(lang, _dict2phonecode[dicttype])
    else:
        newlex = lex(lang, dicttype)
    return newlex.read(filename, lang, dicttype, params)



from typing import List, Union, Optional

import numpy as np
import pandas as pd

from .features import Phone
from . import PhoneCollection

try:
    import plotly.express as px
    import plotly.io as pio
    from sklearn.decomposition import PCA
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        'when using plots, make sure to install the optional "plot" dependency'
    )

# def set_renderer(renderer):
#     pio.renderers.default = renderer

# set_renderer("jupyterlab")


def plot_phone(phone):
    fig = px.bar(x=phone.feature_names, y=phone.vector)
    fig.show()


Phone.plot = plot_phone


def plot_collection(collection, n_components: int = 2, **kwargs):
    vectors = []
    phones = []
    langs = []

    if n_components not in [2, 3, 4]:
        raise ValueError("n_components must be either 2, 3 or 4")

    for phone in collection.values:
        vectors.append(phone.vector)
        langs.append(phone.language_code)
        phones.append(phone.index)
    pca = PCA(n_components=n_components)
    vectors = pca.fit_transform(np.array(vectors))
    if n_components == 2:
        df = pd.DataFrame(
            {
                "phone": phones,
                "language": langs,
                "PC1": vectors[:, 0],
                "PC2": vectors[:, 1],
            }
        )
    elif n_components == 3:
        df = pd.DataFrame(
            {
                "phone": phones,
                "language": langs,
                "PC1": vectors[:, 0],
                "PC2": vectors[:, 1],
                "PC3": vectors[:, 2],
            }
        )
    else:
        df = pd.DataFrame(
            {
                "phone": phones,
                "language": langs,
                "PC1": vectors[:, 0],
                "PC2": vectors[:, 1],
                "PC3": vectors[:, 2],
                "PC4": ((vectors[:, 3]-vectors[:, 3].min())/(vectors[:, 3].max()-vectors[:, 3].min()))*20+1,
            }
        )
    if collection.source.language_column is None:
        if n_components == 2:
            fig = px.scatter(
                df,
                x="PC1",
                y="PC2",
                color="language",
                text="phone",
                title="PCA of Phones",
                **kwargs,
            )
        elif n_components == 3:
            fig = px.scatter_3d(
                df,
                x="PC1",
                y="PC2",
                z="PC3",
                color="language",
                text="phone",
                title="PCA of Phones",
                **kwargs,
            )
        else:
            fig = px.scatter_3d(
                df,
                x="PC1",
                y="PC2",
                z="PC3",
                color="language",
                text="phone",
                title="PCA of Phones",
                **kwargs,
            )
    else:
        new_langs = []
        for i, row in df.iterrows():
            if n_components == 2:
                shared = df[(df["PC1"] == row["PC1"]) & (df["PC2"] == row["PC2"])]
            elif n_components == 3:
                shared = df[
                    (df["PC1"] == row["PC1"])
                    & (df["PC2"] == row["PC2"])
                    & (df["PC3"] == row["PC3"])
                ]
            else:
                shared = df[
                    (df["PC1"] == row["PC1"])
                    & (df["PC2"] == row["PC2"])
                    & (df["PC3"] == row["PC3"])
                    & (df["PC4"] == row["PC4"])
                ]
            
            if len(shared) > 1:
                unique_langs = pd.unique(shared["language"])
                new_langs.append(",".join(sorted(unique_langs)))
            else:
                new_langs.append(row["language"])
        df["language"] = new_langs
        df = df.drop_duplicates()
        if n_components == 2:
            fig = px.scatter(
                df,
                x="PC1",
                y="PC2",
                color="language",
                text="phone",
                title="PCA of Phones",
                **kwargs,
            )
        elif n_components == 3:
            fig = px.scatter_3d(
                df,
                x="PC1",
                y="PC2",
                z="PC3",
                color="language",
                text="phone",
                title="PCA of Phones",
                **kwargs,
            )
        else:
            fig = px.scatter_3d(
                df,
                x="PC1",
                y="PC2",
                z="PC3",
                size="PC4",
                color="language",
                text="phone",
                title="PCA of Phones",
                **kwargs,
            )
    fig.update_traces(textposition="bottom right")
    fig.show()


PhoneCollection.plot = plot_collection



sents = {
    "eng": {
        "word": "THIS IS A TEST",
        "arpabet": "DH IH S IH Z AH0 T EH1 S T",
        "ipa": "ð ɪ s ɪ z ə t ˈɛ s t",
        "disc": "D I s I z @ t 'E s t",
        "xsampa": 'D I s I z @ t "E s t',
    },
    "amh": {
        "word": "ይህ ፈተና ነው",
        "ipa": "j ɨ h f ə . t ə . n a n ə w",
        "xsampa": "j 1 h f @ . t @ . n a n @ w",
    },
    "ben": {
        "word": "এটা একটা পরীক্ষা",
        "ipa": "e . ʈ a æ k . ʈ a p o . r i k . kʰ a",
        "xsampa": "e . t` a { k . t` a p o . r i k . k_h a",
    },
    "deu": {
        "word": "dies ist ein Test",
        "disc": "d 'i s 'I s t 'W n t 'E s t",
        "xsampa": 'd "i: s "I s t "ai n t "E s t',
        "ipa": "d ˈiː s ˈɪ s t ˈai n t ˈɛ s t",
    },
    "nld": {
        "word": "dit is een test",
        "disc": "d 'I t 'I s 'e n t 'E s t",
        "xsampa": 'd "I t "I s "e: n t "E s t',
        "ipa": "d ˈɪ t ˈɪ s ˈeː n t ˈɛ s t",
    },
    "cmn": {
        "word": "这 是 一个 测试",
        "callhome": "Z e4 S %4 i2 g &4 c &4 S %4",
        "xsampa": "t`s` e_1_5 s` M_1_5 i_3_1 k @_1_5 ts_h @_1_5 s` M_1_5",
        "ipa": "ʈʂ e˥˩ ʂ ɯ˥˩ i˧˥ k ə˥˩ tsʰ ə˥˩ ʂ ɯ˥˩",
    },
    "spa": {
        "word": "esto es una prueba",
        "callhome": "e1 s t o e1 s u1 n a p r w e1 B a",
        "xsampa": '"e s t o "e s "u n a p 4 w "e B a',
        "ipa": "ˈe s t o ˈe s ˈu n a p ɾ w ˈe β a",
    },
    "arz": {
        "word": "هذا اختبار",
        "callhome": "h a d A A1 x t i b A2 r",
        "xsampa": 'h a d a: "a: x t i b %a: 4',
        "ipa": "h a d aː ˈaː x t i b ˌaː ɾ",
    },
}



"""
Sources for phone features.

## ``sources.PHOIBLE``

Source for phone inventories from [phoible.org](https://phoible.org).
Use the following citation:

```
@article{moran2014phoible,
title={PHOIBLE online},
author={Moran, Steven and McCloy, Daniel and Wright, Richard},
year={2014},
publisher={Max Planck Institute for Evolutionary Anthropology}
}
```

## ``sources.PANPHON``

Source for phone inventories from [panphon](https://github.com/dmort27/panphon).
Use the following citation:

```
@inproceedings{mortensen2016panphon,
title={Panphon: A resource for mapping IPA segments to articulatory feature vectors},
author={Mortensen, David R and Littell, Patrick and Bharadwaj, Akash and Goyal, Kartik and Dyer, Chris and Levin, Lori},
booktitle={Proceedings of COLING 2016, the 26th International Conference on Computational Linguistics: Technical Papers},
pages={3475--3484},
year={2016}
}
```
"""

from dataclasses import dataclass
from typing import List


@dataclass
class PhoneSource:
    """
    The PhoneSource class is a dataclass that stores the information about a phone source.

    A phone source is a csv containing phone definitions and their linguistic features

    Attributes:
        urls: a list of urls to the csv files
        index_column: the name of the column that contains the index (ipa character(s)) of the phone
        feature_columns: a list of the names of the columns that contain the features of the phone
        language_column: the name of the column that contains the iso code of the language
        dialect_column: the name of the column that contains the dialect
    """

    urls: List[str]
    index_column: str
    feature_columns: List[str]
    language_column: str = None
    allophone_column: str = None
    dialect_column: str = None


PHOIBLE = PhoneSource(
    ["https://raw.githubusercontent.com/phoible/dev/master/data/phoible.csv"],
    "Phoneme",
    [
        "tone",
        "stress",
        "syllabic",
        "short",
        "long",
        "consonantal",
        "sonorant",
        "continuant",
        "delayedRelease",
        "approximant",
        "tap",
        "trill",
        "nasal",
        "lateral",
        "labial",
        "round",
        "labiodental",
        "coronal",
        "anterior",
        "distributed",
        "strident",
        "dorsal",
        "high",
        "low",
        "front",
        "back",
        "tense",
        "retractedTongueRoot",
        "advancedTongueRoot",
        "periodicGlottalSource",
        "epilaryngealSource",
        "spreadGlottis",
        "constrictedGlottis",
        "fortis",
        "raisedLarynxEjective",
        "loweredLarynxImplosive",
        "click",
    ],
    "ISO6393",
    "Allophones",
    "SpecificDialect",
)

PANPHON = PhoneSource(
    [
        "https://raw.githubusercontent.com/dmort27/panphon/master/panphon/data/ipa_all.csv",
        "https://raw.githubusercontent.com/dmort27/panphon/master/panphon/data/ipa_bases.csv",
    ],
    "ipa",
    [
        "syl",
        "son",
        "cons",
        "cont",
        "delrel",
        "lat",
        "nas",
        "strid",
        "voi",
        "sg",
        "cg",
        "ant",
        "cor",
        "distr",
        "lab",
        "hi",
        "lo",
        "back",
        "round",
        "velaric",
        "tense",
        "long",
        "hitone",
        "hireg",
    ],
)



"""
This module allows to convert between the "ipa", "xsampa" and "arpabet" formats.
The code is adapted from the [phonecodes package](https://github.com/jhasegaw/phonecodes) by Mark Hasegawa-Johnson.

Examples:
    A converter object can be used.
    ```py
    from phone.convert import Converter
    converter = Converter()
    converter("wɜ˞ld", "ipa", "arpabet")
    ```
    > ``['W', 'ER', 'L', 'D']``

    You can also list all possible formats.
    ```py
    converter.formats
    ```
"""
from typing import List, Optional
import re

from .phonecodes.src import phonecodes
from phones.normalize import normalize_unicode

class Converter:
    def __init__(self) -> None:
        self.phonecodes = {
            ("arpabet", "ipa"): (phonecodes.arpabet2ipa, False),
            ("ipa", "arpabet"): (phonecodes.ipa2arpabet, False),
            ("ipa", "callhome"): (phonecodes.ipa2callhome, ["arz", "cmn", "spa"]),
            ("callhome", "ipa"): (phonecodes.callhome2ipa, ["arz", "cmn", "spa"]),
            ("ipa", "disc"): (phonecodes.ipa2disc, False),
            ("disc", "ipa"): (phonecodes.disc2ipa, ["nld", "eng", "deu"]),
            ("ipa", "xsampa"): (phonecodes.ipa2xsampa, False),
            ("xsampa", "ipa"): (phonecodes.xsampa2ipa, False),
            ("arpabet", "xsampa"): (self.arpabet2xsampa, False),
            ("xsampa", "arpabet"): (self.xsampa2arpabet, False),
        }

    def xsampa2arpabet(self, x, lang):
        if not lang:
            lang = None
        return self(
            self(
                x,
                "xsampa",
                "ipa",
                lang,
                True
            ),
            "ipa",
            "arpabet",
            lang, 
            True
        )

    def arpabet2xsampa(self, x, lang):
        if not lang:
            lang = None
        return self(
            self(
                x,
                "arpabet",
                "ipa",
                lang,
                True
            ),
            "ipa",
            "xsampa",
            lang,
            True
        )

    def __call__(
        self, x, _from: str, _to: str = "ipa", lang: Optional[str] = None, return_str = False,
    ) -> List[str]:
        """
        It takes a string, and replaces all the symbols of the ``_from`` format to the ``_to`` format.

        Args:
            x: the string to be converted
            _from: the name of the format to convert from
            _to: the name of the format to convert to

        Returns:
            The converted string.
        """
        func, langs = self.phonecodes[(_from, _to)]
        assert not (not langs and lang is not None)
        result = func(normalize_unicode(x), langs)
        result = [normalize_unicode(r) for r in result]
        if max([len(r) for r in result]) > 1:
            result_new = []
            for r in result:
                result_new += [r, ' ']
            result = result_new[:-1]
        if return_str:
            result = "".join(result)
            result = re.sub(r"\s+", " ", result).strip()
        else:
            result = [r for r in result if len(r.replace(" ", "")) > 0]
        return result

    @property
    def formats(self) -> List[str]:
        return list(set([x[0] for x in self.phonecodes.keys()]))



from functools import lru_cache
import hashlib
from lib2to3.pgen2.token import OP
import os
from typing import Callable, Dict, Iterable, List, Tuple, Union
import unicodedata
import pandas as pd
import numpy as np
from tqdm.auto import tqdm
from scipy.spatial import distance
from copy import deepcopy
from pathlib import Path
import re

from .sources import PhoneSource, PHOIBLE
from .features import Phone


class PhoneCollection:
    def __init__(
        self,
        source: PhoneSource = PHOIBLE,
        cache_dir: str = f"{str(Path.home())}/.cache/phones",
        merge_same_language: bool = True,
        load_dialects: bool = False,
        _master: object = None,
    ) -> None:
        """Creates a ``PhoneCollection`` object that loads phones from a ``PhoneSource`` into a pandas DataFrame.

        Args:
            source: The ``PhoneSource`` object that defines the source of the data.
            cache_dir: The directory where the data will be downloaded and cached.
            merge_same_language: If true, multiple phone definitions in the same language are merged.
            load_dialects: If false, dialects are ignored.
        """
        self.source = source
        self.source.feature_columns = sorted(self.source.feature_columns)
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir)
        dfs = []
        for url in source.urls:
            url_hash = hashlib.sha224(url.encode()).hexdigest()
            download_path = os.path.join(cache_dir, url_hash + ".pkl")
            if os.path.isfile(download_path):
                df = pd.read_pickle(download_path)
            else:
                df = pd.read_csv(url, dtype=str)
                df.to_pickle(download_path)
            dfs.append(df)
        self.data = pd.concat(dfs)

        for feature in self.source.feature_columns:
            self.data[feature] = self.data[feature].apply(
                PhoneCollection.feature_to_weight
            )
        self.columns = [
            self.source.index_column,
            self.source.language_column,
            self.source.allophone_column,
        ]
        if not load_dialects and self.source.dialect_column is not None:
            self.data = self.data[self.data[self.source.dialect_column].isna()]
        else:
            self.columns.append(self.source.dialect_column)
        self.columns = [c for c in self.columns if c is not None]

        if merge_same_language and self.source.language_column is not None:
            self.data = self.data.dropna(subset=[self.source.language_column])

            self.data = (
                self.data.groupby(self.columns, dropna=False)[
                    self.source.feature_columns
                ]
                .mean()
                .reset_index()
            )

        if self.source.language_column is not None:
            self.data = self.data.dropna(
                subset=[self.source.index_column, self.source.language_column]
            )
        else:
            self.data = self.data.dropna(subset=[self.source.index_column])

        self.data[self.source.index_column] = self.data[self.source.index_column].apply(
            lambda x: unicodedata.normalize("NFC", x)
        )
        if self.source.allophone_column is not None:
            self.data[self.source.allophone_column] = self.data[
                self.source.allophone_column
            ].apply(lambda x: unicodedata.normalize("NFC", str(x)))
        self.data = self.data[self.columns + self.source.feature_columns]
        self._master = deepcopy(self)
        self.lang_filter = None
        self.load_dialects = load_dialects

    @property
    def features(self):
        return self.source.feature_columns

    @staticmethod
    def feature_to_weight(feature: str) -> float:
        """
        If the feature is a string, try to convert it to a float "-" is converted to -1, "+" to 1.
        If it's a string but can't be converted to a float, return 0.0.
        If it's a comma-delimited list of "+" and "-", return the mean of the list of floats.

        Args:
            feature: The feature to be converted to a weight.

        Return:
            The string feature converted to a float in `[-1.,0.,1.]`
        """
        if isinstance(feature, str):
            try:
                return float(feature.replace("-", "-1").replace("+", "1"))
            except ValueError:
                if "," in feature:
                    return np.mean(
                        [
                            float(f.replace("-", "-1").replace("+", "1"))
                            for f in feature.split(",")
                        ]
                    )
                else:
                    return 0.0

    @property
    def lang_list(self) -> List[str]:
        return list(sorted(self.data[self.source.language_column].unique()))

    @property
    def phone_list(self) -> List[str]:
        return list(sorted(self.data[self.source.index_column].unique()))

    @property
    def dialect_list(self) -> List[str]:
        if not self.load_dialects:
            raise ValueError("Dialects are not loaded.")
        return list(sorted(self.data[self.source.dialect_column].dropna().unique()))

    def phones(self, phones: Union[str, List[str]]) -> object:
        """
        It takes a list of phones and returns a copy ``PhoneCollection`` with only the rows that have one of
        those phones.

        Args:
            phones: A list of phones or single phone to filter on.

        Returns:
            A new instance of the class, with the filtered data.
        """
        _self = deepcopy(self)
        if len(phones) > 0:
            if not isinstance(phones, list):
                phones = [phones]
            phones = [unicodedata.normalize("NFC", p) for p in phones]
            _self.data = _self.data[_self.data[self.source.index_column].isin(phones)]
        _self._master = deepcopy(self)
        return _self

    def dialects(self, dialects: Union[str, List[str], None], inplace=True) -> object:
        """
        It takes a list of dialects and returns a copy ``PhoneCollection`` with only the rows that have one of
        those dialects.

        Args:
            dialects: A list of dialects or single dialects to filter on. Use ```None``` to remove all dialects except the one without a specific name.
            inplace: Modifies the underlying dataframe, affecting phones.

        Returns:
            A new instance of the class, with the filtered data.
        """
        if not self.load_dialects:
            raise ValueError("Dialects are not loaded.")
        _self = deepcopy(self)
        if len(dialects) > 0:
            if not isinstance(dialects, list):
                dialects = [dialects]
            dialect_mask = _self.data[self.source.dialect_column].str.match(
                "|".join([re.escape(d).lower() for d in dialects]), case=False
            )
            if dialects != [None]:
                dialect_mask = dialect_mask.fillna(False)  # remove standard dialect
            else:
                dialect_mask = dialect_mask.isna()
            _self.data = _self.data[dialect_mask]
        _self.dialect_filter = set(dialects)
        if inplace:
            _self._master.data = _self.data
            _self._master.dialect_filter = _self.dialect_filter
        return _self

    def langs(self, langs, inplace=True) -> object:
        """
        It takes a list of languages and returns a copy ``PhoneCollection`` with only the rows that have one of
        those languages.

        Args:
            langs: A list of languages or single language to filter on.
            inplace: Modifies the underlying dataframe, affecting phones.

        Returns:
            A new instance of the class, with the filtered data.
        """
        _self = deepcopy(self)
        if len(langs) > 0:
            if not isinstance(langs, list):
                langs = [langs]
            _self.data = _self.data[_self.data[self.source.language_column].isin(langs)]
        _self.lang_filter = set(langs)
        if inplace:
            _self._master.data = _self.data
            _self._master.lang_filter = _self.lang_filter
        return _self

    @property
    def values(self) -> List[object]:
        """
        The collection as a list of phones.

        Returns:
            A list of ``Phone`` objects.
        """
        phone_df = (
            self.data.groupby(
                [c for c in self.columns if c != self.source.allophone_column]
            )
            .mean(numeric_only=True)
            .reset_index()
        )
        return [self._row_to_phone(row) for _, row in phone_df.iterrows()]

    @property
    def values_with_allophones(self) -> List[object]:
        """
        The collection as a list of phones.

        Returns:
            A list of ``Phone`` objects.
        """
        phone_df = self.data.groupby([c for c in self.columns]).mean().reset_index()
        return [self._row_to_phone(row) for _, row in phone_df.iterrows()]

    @property
    def val(self) -> object:
        """
        If the collection is filtered down to a single phone, return that phone.

        Returns:
            A ``Phone`` object.
        """
        phone_df = self.data.groupby(self.columns).mean().reset_index()
        if len(phone_df) > 1:
            phone_df = self.data.groupby(self.source.index_column).mean().reset_index()
        results = [self._row_to_phone(row) for _, row in phone_df.iterrows()]
        assert len(results) == 1
        return results[0]

    def _get_phone_inventory(self, language: str):
        phone_df = self.data.groupby(self.columns).mean().reset_index()
        phone_df = phone_df[phone_df[self.source.language_column] == language]
        return phone_df

    def get_mean_allophone_distance(
        self, distance_weights=None, show_progress=False
    ) -> float:
        """
        For each row in the dataframe, we get the phone and allophone values.
        If the allophone is different from the phone, we get the mean distance between the allophone
        and the phone. We return the mean of all allophone <-> phone distances.

        Args:
            distance_weights: A dictionary of weights for each distance type.
            show_progress: If True, show a progress bar.

        Returns:
            The mean of the distances between allophones and their phones.
        """
        dists = []
        for _, row in tqdm(
            self.data.iterrows(), total=len(self.data), disable=(not show_progress)
        ):
            phone = row[self.source.index_column]
            for allophone in row[self.source.allophone_column].split():
                if allophone != phone:
                    try:
                        dists.append(
                            self.get_mean_phone_distance(
                                allophone, phone, distance_weights=distance_weights
                            )
                        )
                    except ValueError:
                        for subphone in allophone:
                            try:
                                dists.append(
                                    self.get_mean_phone_distance(
                                        subphone,
                                        phone,
                                        distance_weights=distance_weights,
                                    )
                                )
                            except ValueError:
                                pass
        return np.mean(dists)

    @lru_cache(maxsize=1024)
    def get_mean_phone_distance(
        self,
        phone: str,
        other_phone: str,
        distance_fn: Callable[
            [Iterable[float], Iterable[float]], float
        ] = distance.euclidean,
        distance_weights=None,
    ) -> float:
        """
        For a given phone, find the mean of all the features for that phone. Then, find the
        distance between that phone and another phone.

        Args:
            phone: The phone to compare to the other phone.
            other_phone: The other phone to compare to.
            distance_fn: The distance function to use.
            distance_weights: This is a list of weights for each feature.

        Return:
            The mean distance between the two phones.
        """
        phones_df = self.data.groupby(self.source.index_column).mean().reset_index()
        phone1_df = phones_df[phones_df[self.source.index_column] == phone][
            self.source.feature_columns
        ].values.flatten()
        phone2_df = phones_df[phones_df[self.source.index_column] == other_phone][
            self.source.feature_columns
        ].values.flatten()
        if distance_weights is None:
            return distance_fn(phone1_df, phone2_df)
        else:
            distance_weights = np.array(distance_weights)
            distance_weights = (
                distance_weights / distance_weights.sum() * len(distance_weights)
            )
            return distance_fn(
                phone1_df * distance_weights, phone2_df * distance_weights
            )

    def get_closest_by_vector(
        self,
        vector: List[float],
        distance_fn: Callable[
            [Iterable[float], Iterable[float]], float
        ] = distance.euclidean,
    ) -> List[Tuple[float, object]]:
        """Given a vector, find the phone that is closest to the vector

        Args:
            vector: The vector we're looking for the closest phones to.
            distance_fn: The function that will be used to calculate the distance between the vector and phones.

        Returns:
            A list of tuples, where each tuple contains a distance and a phone.
        """
        phones_df = (
            self.data.groupby(
                [c for c in self.columns if c != self.source.allophone_column]
            )
            .mean()
            .reset_index()
        )
        phones_df.drop_duplicates(inplace=True)
        smallest_dist = float("inf")
        results = []
        for _, row in phones_df.iterrows():
            dist = distance_fn(vector, row[self.source.feature_columns].values)
            if dist <= smallest_dist:
                if dist < smallest_dist:
                    results = []
                    smallest_dist = dist
                results.append((dist, self._row_to_phone(row)))
        return results

    def get_closest_by_phone(
        self,
        phone: List[float],
        distance_fn: Callable[
            [Iterable[float], Iterable[float]], float
        ] = distance.euclidean,
    ) -> List[Tuple[float, object]]:
        """Given a phone, return the closest phone in the collection

        Args:
            phone: The phone to find the closest phone to.
            distance_fn: The function that will be used to measure the distance between phones.

        Returns:
            A list of tuples, where each tuple contains a distance and a phone.
        """
        return self.get_closest_by_vector(phone.vector, distance_fn)

    def get_closest(
        self,
        phone: str,
        src_language: str,
        tgt_language: str,
        return_allophones: bool = False,
        distance_fn: Callable[
            [Iterable[float], Iterable[float]], float
        ] = distance.euclidean,
        distance_weights=None,
        allow_allophones=True,
        return_all=False,
    ) -> Union[
        List[Tuple[float, str]], Tuple[List[Tuple[float, str]], List[Tuple[float, str]]]
    ]:
        """Given a phone, a source language, a target language, a distance function, and a distance weight,
        `get_closest` returns the closest phone in the target language.
        It also returns all the allophones of the source phone in the target language.
        It also returns the distance between the source phone and the closest phone.

        Example:
        Let's say we want to find the closest phone to the phone `ð` in the language `English` in `German`.
        ```py
        pc = PhoneCollection()
        pc.get_closest("ð", "eng", "deu")
        ```
        > ``[(2.8284271247461903, 'z'), (2.8284271247461903, 'ʒ')]``

        Args:
            phone: The phone to be mapped.
            src_language: The language of the phone that you want to find the closest one to.
            tgt_language: The language of the target phone.
            return_allophones: If True, return a tuple of ``(closest_phones, allophones)``
            distance_fn: The distance function to use.
            distance_weights:
                If None, the distance weights are set to 1/n, where n is the number of features.
                Otherwise, the weights are normalised and then used for the distance calculations.
            allow_allophones: If True, then if the phone is not found in the inventory, search for a phone the given ``phone`` is an allophone of.
            return_all: If True, return all phones and their distances, not just the closest ones.

        Returns:
            Returns a list of ``(distance, phone)`` for the closests phones or for all phones if ``return_all`` is True.
            If ``allow_allophones`` is True, returns a Tuple of lists with the first entry being the closests phones and the second being allophones.
        """
        assert self.source.language_column is not None
        phone = unicodedata.normalize("NFC", phone)
        src_phone_df = self._get_phone_inventory(src_language)
        src_phone_vec = src_phone_df[src_phone_df[self.source.index_column] == phone][
            self.source.feature_columns
        ]
        if len(src_phone_vec) == 0 and allow_allophones:
            src_phone_vec = src_phone_df[
                src_phone_df[self.source.allophone_column].apply(
                    lambda x: phone in x.split()
                )
            ][self.source.feature_columns]
        if len(src_phone_vec) > 0:
            src_phone_vec = np.mean(src_phone_vec, axis=0)
        tgt_phone_df = self._get_phone_inventory(tgt_language)
        smallest_dist = float("inf")
        results = []
        allophones = []
        for tgt_phone in tgt_phone_df[self.source.index_column].unique():
            rows = tgt_phone_df[tgt_phone_df[self.source.index_column] == tgt_phone]
            tgt_phone_vec = rows[self.source.feature_columns].values
            if len(tgt_phone_vec) > 0:
                tgt_phone_vec = np.mean(tgt_phone_vec, axis=0)
            if distance_weights is None:
                distance_weights = np.ones(len(self.source.feature_columns))
                distance_weights = (
                    distance_weights / distance_weights.sum() * len(distance_weights)
                )
            dist = distance_fn(
                src_phone_vec * distance_weights, tgt_phone_vec * distance_weights
            )
            for _, row in rows.iterrows():
                if phone in row[self.source.allophone_column].split():
                    p = self._row_to_phone(row)
                    p.index = phone
                    allophones.append((dist, p))
                    break
            if dist <= smallest_dist or return_all:
                if dist < smallest_dist:
                    smallest_dist = dist
                    if not return_all:
                        results = []
                results.append((dist, self._row_to_phone(row)))

        if return_allophones:
            return sorted(results), allophones
        else:
            return sorted(results)

    def _row_to_phone(self, row):
        idx = row[self.source.index_column]
        features = {f: row[f] for f in self.source.feature_columns if row[f] != "N"}
        language = None
        if self.source.language_column in row:
            language = row[self.source.language_column]
        allophones = None
        if self.source.allophone_column in row:
            allophones = row[self.source.allophone_column]
        if len(features) > 0:
            try:
                _master = self._master
            except AttributeError:
                _master = self
            return Phone(idx, features, language, allophones, _master)
        return None



import sys, os, re
import pprint
import filecmp

sys.path.append("../src")
import pronlex
import phonecodes

"""This script should be run from the command line, in the test directory.
   It will load some pronlexes from the 'fixtures' subdirectory,
   test phone code conversion, and test both word and phone searches."""

os.makedirs("outputs", exist_ok=True)

fixtures = [
    ["isle_eng_sample", ["arpabet", "disc", "xsampa"]],
    ["babel_amh_sample", ["xsampa"]],
    ["babel_ben_sample", ["xsampa"]],
    ["celex_deu_sample", ["disc", "xsampa"]],
    ["celex_eng_sample", ["arpabet", "disc", "xsampa"]],
    ["celex_nld_sample", ["disc", "xsampa"]],
    ["callhome_arz_sample", ["callhome", "xsampa"]],
    ["callhome_cmn_sample", ["callhome", "xsampa"]],
    ["callhome_spa_sample", ["callhome", "xsampa"]],
]
p = {}
original = sys.stdout
sys.stdout = open(os.path.join("outputs", "test_pronlex.txt"), "w")

#########################################################################
# Test reading in dictionaries, converting phonecodes, and writing them
for fixture in fixtures:
    srcfile = fixture[0]
    p[srcfile] = {}
    (dtype, lang, rem) = srcfile.split("_")
    dict_params = {}
    if dtype == "isle":
        dict_params["discard_phones"] = "#."
    print("Reading %s dict in %s from %s" % (dtype, lang, srcfile))
    p[srcfile]["ipa"] = pronlex.read(
        os.path.join("fixtures", srcfile) + ".txt", lang, dtype, dict_params
    ).recode("ipa")

    # Convert to each target code, and back again, and check results
    for c1 in fixture[1]:
        for c in [["ipa", c1], [c1, "ipa"]]:
            print("##########################################################")
            print("# Testing %s[%s].recode(%s)" % (srcfile, c[0], c[1]))
            p[srcfile][c[1]] = p[srcfile][c[0]].recode(c[1])
            destfile = re.sub(r"sample", c[1], srcfile)
            print("%s -> %s" % (srcfile, destfile))
            p[srcfile][c[1]].save(os.path.join("outputs", destfile) + ".txt")
            with open(os.path.join("fixtures", destfile) + ".txt") as f:
                flines = f.readlines()
                flines.append("-----\n")
            with open(os.path.join("outputs", destfile) + ".txt") as f:
                olines = f.readlines()
                olines.append("-----\n")
            if flines == olines:
                print("Pass")
            else:
                print("".join(["FAIL\n"] + flines + olines))

####################################################################
# Test looking up words
from sentences import sents

slists = {
    L: {c: [p for p in S.split(" ")] for (c, S) in D.items()}
    for (L, D) in sents.items()
}
for fixture in fixtures:
    srcfile = fixture[0]
    (dtype, lang, rem) = srcfile.split("_")
    for c1 in ["ipa"] + fixture[1]:
        # Test words2phones
        print("##########################################################")
        print("# Testing words2phones(%s,%s)" % (srcfile, c1))

        res = p[srcfile][c1].words2phones(slists[lang]["word"])
        pat = re.compile(" ".join(slists[lang][c1]), re.IGNORECASE)
        if res == slists[lang][c1]:
            print("Pass")
        else:
            print("FAIL")
            m = min(len(res), len(slists[lang][c1]))
            for n in range(0, m):
                if slists[lang][c1][n] == res[n]:
                    print("%s == %s" % (res[n], slists[lang][c1][n]))
                else:
                    print("%s != %s" % (res[n], slists[lang][c1][n]))
            if len(slists[lang][c1]) > m:
                print("Ref chars not in hyp:" + ":".join(slists[lang][c1][m:]))
            elif len(res) > m:
                print("Hyp chars not in ref:" + ":".join(res[m:]))

        # Test phones2words
        print("##########################################################")
        print("# Testing phones2words(%s,%s)" % (srcfile, c1))
        res = p[srcfile][c1].phones2words(slists[lang][c1])
        pat = re.compile(" ".join(slists[lang]["word"]), re.IGNORECASE)
        if any(re.match(pat, " ".join(x)) for x in res[0]):
            print("Pass")
        else:
            print("FAIL")
            print("Target:" + ":".join(slists[lang]["word"]))
            print("Results:[" + "][".join([":".join(x) for x in res]) + "]")


# Test phones2words with 1 or 2 phone distance allowed
print("\n##########################################################")
print("# Testing phones2words(isle_eng_sample) with dist==2")
srcfile = "isle_eng_sample"
lang = "eng"
c1 = "ipa"
res = p[srcfile][c1].phones2words(slists[lang][c1], 2)
for d in sorted(res.keys()):
    print("### Candidates with string edit distance == %d" % (d))
    for c in res[d]:
        print(" ".join(c))



from phones.convert import Converter
from phones import PhoneCollection

converter = Converter()


# The following tests are from
# https://github.com/rossellhayes/ipa/blob/main/tests/testthat/test-ipa.R
def test_convert_xsampa_ipa():
    assert "ˈnom.bɾe" == converter('"nom.b4e', "xsampa", "ipa", return_str=True)
    assert "nɔ̃bʁ" == converter("nO~bR", "xsampa", "ipa", return_str=True)
    assert "ˌhɛˈloʊ" == converter('%hE"loU', "xsampa", "ipa", return_str=True)
    assert "ˌfoʊ.naɪˈif" == converter('/%foU.naI"if/', "xsampa", "ipa", return_str=True)
    assert "wɜ˞ld" == converter("w3`ld", "xsampa", "ipa", return_str=True)


def test_convert_ipa_xsampa():
    assert '"nom.b4e' == converter("ˈnom.bɾe", "ipa", "xsampa", return_str=True)
    assert "nO~bR" == converter("nɔ̃bʁ", "ipa", "xsampa", return_str=True)
    assert '%hE"loU' == converter("ˌhɛˈloʊ", "ipa", "xsampa", return_str=True)
    assert '%foU.naI"if' == converter("ˌfoʊ.naɪˈif", "ipa", "xsampa", return_str=True)
    assert "w3`ld" == converter("wɜ˞ld", "ipa", "xsampa", return_str=True)


def test_convert_arpa_ipa():
    assert "ɑ ɹ p ʌ" == converter("AA R P AH", "arpabet", "ipa", return_str=True)
    assert "h ɛ l oʊ" == converter("HH EH L OW", "arpabet", "ipa", return_str=True)
    assert "w ɜ˞ l d" == converter("W ER L D", "arpabet", "ipa", return_str=True)


def test_convert_ipa_arpa():
    assert "AA R P AH" == converter("ɑɹpʌ", "ipa", "arpabet", return_str=True)
    assert "HH EH L OW" == converter("hɛloʊ", "ipa", "arpabet", return_str=True)
    assert "W ER L D" == converter("w ɜ˞ l d", "ipa", "arpabet", return_str=True)


def test_convert_arpa_xsampa():
    assert "A r\\ p V" == converter("AA R P AH", "arpabet", "xsampa", return_str=True)
    assert "h E l oU" == converter("HH EH L OW", "arpabet", "xsampa", return_str=True)
    assert "w 3` l d" == converter("W ER L D", "arpabet", "xsampa", return_str=True)


def test_convert_xsampa_arpa():
    assert "AA R P AH" == converter("A r\\ p V", "xsampa", "arpabet", return_str=True)
    assert "HH EH L OW" == converter("h E l oU", "xsampa", "arpabet", return_str=True)
    assert "W ER L D" == converter("w 3` l d", "xsampa", "arpabet", return_str=True)


def test_convert_normalization():
    assert "W ER L D" == converter("w ɝ l d", "ipa", "arpabet", return_str=True)
    assert "W ER0 L D" == converter("w ɚ l d", "ipa", "arpabet", return_str=True)


def test_phone_collection():
    pc = PhoneCollection()
    ph = pc.langs("eng").values[0]
    assert str(ph) == "aː (eng)"



#!/usr/bin/python3
"""A set of convenience functions for converting among different phone codes.
   Usage: 
   import phonecodes
   print phonecodes.CODES   # the known phone codes
   print phonecodes.LANGUAGES # the known languages
   s1 = phonecodes.convert(s0, code0, code1, language)
   # s0 and s1 are strings containing individual symbols
   # code0 and code1 must be members of phonecodes.CODES, of course
   # language must be a member of phonecodes.LANGUAGES, of course
   #   (but not all languages are known for all phone codes)
   l1 = phonecodes.convertlist(l0, code0, code1, language)
   # l0, l1 are lists of symbols
   phonecodes.vowels
   phonecodes.consonants
   # list known IPA symbols of vowels, consonants.
   # for other tables, see phonecode_tables.py
"""
import re, sys
from . import phonecode_tables

CODES = set(("ipa", "arpabet", "xsampa", "disc", "callhome"))
LANGUAGES = set(("eng", "deu", "nld", "arz", "cmn", "spa", "yue", "lao", "vie"))

vowels = phonecode_tables._ipa_vowels
consonants = phonecode_tables._ipa_consonants

#####################################################################
def translate_string(s, d):
    """(tl,ttf)=translate_string(s,d):
    Translate the string, s, using symbols from dict, d, as:
    1. Min # untranslatable symbols, then 2. Min # symbols.
    tl = list of translated or untranslated symbols.
    ttf[n] = True if tl[n] was translated, else ttf[n]=False."""
    N = len(s)
    symcost = 1  # path cost per translated symbol
    oovcost = 10  # path cost per untranslatable symbol
    maxsym = max(len(k) for k in d.keys())  # max input symbol length
    # (pathcost to s[(n-m):n], n-m, translation[s[(n-m):m]], True/False)
    lattice = [(0, 0, "", True)]
    for n in range(1, N + 1):
        # Initialize on the assumption that s[n-1] is untranslatable
        lattice.append((oovcost + lattice[n - 1][0], n - 1, s[(n - 1) : n], False))
        # Search for translatable sequences s[(n-m):n], and keep the best
        for m in range(1, min(n + 1, maxsym + 1)):
            if s[(n - m) : n] in d and symcost + lattice[n - m][0] < lattice[n][0]:
                lattice[n] = (
                    symcost + lattice[n - m][0],
                    n - m,
                    d[s[(n - m) : n]],
                    True,
                )
    # Back-trace
    tl = []
    translated = []
    n = N
    while n > 0:
        tl.append(lattice[n][2])
        translated.append(lattice[n][3])
        n = lattice[n][1]
    return (tl[::-1], translated[::-1])


def attach_tones_to_vowels(il, tones, vowels, searchstep, catdir):
    """Return a copy of il, with each tone attached to nearest vowel if any.
    searchstep=1 means search for next vowel, searchstep=-1 means prev vowel.
    catdir>=0 means concatenate after vowel, catdir<0 means cat before vowel.
    Tones are not combined, except those also included in the vowels set.
    """
    ol = il.copy()
    v = 0 if searchstep > 0 else len(ol) - 1
    t = -1
    while 0 <= v and v < len(ol):
        if (ol[v] in vowels or (len(ol[v]) > 1 and ol[v][0] in vowels)) and t >= 0:
            ol[v] = ol[v] + ol[t] if catdir >= 0 else ol[t] + ol[v]
            ol = ol[0:t] + ol[(t + 1) :]  # Remove the tone
            t = -1  # Done with that tone
        if v < len(ol) and ol[v] in tones:
            t = v
        v += searchstep
    return ol


#####################################################################
# X-SAMPA
def ipa2xsampa(x, language):
    """Attempt to return X-SAMPA equivalent of an IPA phone x."""
    (tl, ttf) = translate_string(x, phonecode_tables._ipa2xsampa)
    return tl


def xsampa2ipa(x, language):
    """Return the IPA equivalent of X-SAMPA phone x."""
    (tl, ttf) = translate_string(x, phonecode_tables._xsampa_and_diac2ipa)
    return tl


######################################################################
# Language-dependent lexical tones and stress markers
def tone2ipa(n, language):
    return phonecode_tables._tone2ipa[L][int(n[1:])]


#####################################################################
# DISC, the system used by CELEX
def disc2ipa(x, L):
    """Convert DISC symbol x into IPA, for language L"""
    if L == "nld":
        (tl, ttf) = translate_string(x, phonecode_tables._disc2ipa_dutch)
        return tl
    elif L == "eng":
        (tl, ttf) = translate_string(x, phonecode_tables._disc2ipa_english)
        return tl
    else:
        (tl, ttf) = translate_string(x, phonecode_tables._disc2ipa)
        return tl


def ipa2disc(x, L):
    """Convert IPA symbol x into DISC"""
    (tl, ttf) = translate_string(x, phonecode_tables._ipa2disc)
    return tl


def ipa2disc_old(x, L):
    """Convert IPA symbol x into DISC, for language L"""
    # Convert whole thing if possible; otherwise try prefix+vowel; else quit
    if x in phonecode_tables._ipa2disc:
        return phonecode_tables._ipa2disc[x]
    elif x[0] in phonecode_tables._ipa2disc and x[1:] in phonecode_tables._ipa2disc:
        return phonecode_tables._ipa2disc[x[0]] + phonecode_tables._ipa2disc[x[1:]]
    else:
        raise KeyError("Unknown IPA symbol %s for language %s" % (x, L))


#######################################################################
# Callhome phone codes
def callhome2ipa(x, L):
    """Convert callhome phone symbol x into IPA for language L"""
    (il, ttf) = translate_string(x, phonecode_tables._callhome2ipa[L])
    if L == "arz":
        ol = attach_tones_to_vowels(
            il,
            phonecode_tables._ipa_stressmarkers,
            phonecode_tables._ipa_vowels,
            -1,
            -1,
        )
    elif L == "cmn":
        ol = attach_tones_to_vowels(
            il, phonecode_tables._ipa_tones, phonecode_tables._ipa_vowels, -1, 1
        )
    elif L == "spa":
        ol = attach_tones_to_vowels(
            il,
            phonecode_tables._ipa_stressmarkers,
            phonecode_tables._ipa_vowels,
            -1,
            -1,
        )
    return ol


def ipa2callhome(x, L):
    """Convert IPA symbol x into callhome notation, for language L"""
    (il, ttf) = translate_string(x, phonecode_tables._ipa2callhome[L])
    if L == "arz":
        ol = attach_tones_to_vowels(
            il, "012", phonecode_tables._callhome_vowels["arz"], 1, 1
        )
    elif L == "cmn":
        ol = attach_tones_to_vowels(
            il, "012345", phonecode_tables._callhome_vowels["cmn"], -1, 1
        )
    elif L == "spa":
        ol = attach_tones_to_vowels(
            il, "012", phonecode_tables._callhome_vowels["spa"], 1, 1
        )
    return ol


#########################################################################
# ARPABET and TIMIT
def arpabet2ipa(x, language):
    """Convert ARPABET symbol X to IPA"""
    (il, ttf) = translate_string(x, phonecode_tables._arpabet2ipa)
    ol = attach_tones_to_vowels(
        il, phonecode_tables._ipa_stressmarkers, phonecode_tables._ipa_vowels, -1, -1
    )
    return ol


def ipa2arpabet(x, language):
    """Convert IPA symbols to ARPABET"""
    (il, ttf) = translate_string(x, phonecode_tables._ipa2arpabet)
    ol = attach_tones_to_vowels(il, "012", phonecode_tables._arpabet_vowels, 1, 1)
    return ol


def timit2ipa(x, L):
    """Convert TIMIT phone codes to IPA"""
    x = x.upper()
    (il, ttf) = translate_string(x, phonecode_tables._timit2ipa)
    ol = attach_tones_to_vowels(
        il, phonecode_tables._ipa_stressmarkers, phonecode_tables._ipa_vowels, -1, -1
    )
    return ol


#######################################################################
# phonecodes.convert and phonecodes.convertlist
# are used to convert symbols and lists of symbols, respectively,
# to or from IPA, by calling appropriate other functions.
#
_convertfuncs = {
    "arpabet": (arpabet2ipa, ipa2arpabet),
    "xsampa": (xsampa2ipa, ipa2xsampa),
    "disc": (disc2ipa, ipa2disc),
    "callhome": (callhome2ipa, ipa2callhome),
}


def convert(s0, c0, c1, language):
    if c0 == "ipa" and c1 != "ipa":
        x = _convertfuncs[c1][1](s0, language)
        return x
    elif c0 != "ipa" and c1 == "ipa":
        return _convertfuncs[c0][0](s0, language)
    else:
        raise RuntimeError("must convert to/from ipa, not %s to %s" % (c0, c1))


def convertlist(l0, c0, c1, language):
    return [convert(s0, c0, c1, language) for s0 in l0]
